from netbox.search import SearchIndex, register_search

from . import models


@register_search
class BranchIndex(SearchIndex):
    model = models.Branch
    fields = (
        ('name', 100),
        ('description', 500),
        ('comments', 5000),
    )
    display_attrs = ('description', 'description')



from django.db import connection, migrations
from netbox.plugins import get_plugin_config

from netbox_branching.choices import BranchStatusChoices


def copy_migrations(apps, schema_editor):
    """
    Create a copy of the migrations table in each active branch.
    """
    Branch = apps.get_model('netbox_branching', 'Branch')

    table = 'django_migrations'
    schema_prefix = get_plugin_config('netbox_branching', 'schema_prefix')

    with connection.cursor() as cursor:
        main_table = f'public.{table}'

        for branch in Branch.objects.filter(status=BranchStatusChoices.READY):
            print(f'\n    Copying migrations for branch {branch.name} ({branch.schema_id})...', end='')
            schema_name = f'{schema_prefix}{branch.schema_id}'
            schema_table = f'{schema_name}.{table}'

            # Copy the migrations table to the branch schema
            cursor.execute(f"CREATE TABLE {schema_table} ( LIKE {main_table} INCLUDING INDEXES )")

            # Copy table data
            cursor.execute(f"INSERT INTO {schema_table} SELECT * FROM {main_table}")

            # Designate id as an identity column
            cursor.execute(f"ALTER TABLE {schema_table} ALTER COLUMN id ADD GENERATED BY DEFAULT AS IDENTITY")

            # Set the next value for the ID sequence
            cursor.execute("SELECT MAX(id) from django_migrations")
            starting_id = cursor.fetchone()[0] + 1
            cursor.execute(f"ALTER SEQUENCE {schema_name}.django_migrations_id_seq RESTART WITH {starting_id}")

    print('\n ', end='')  # Padding for final "OK"


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_branching', '0003_rename_indexes'),
    ]

    operations = [
        migrations.RunPython(
            code=copy_migrations,
            reverse_code=migrations.RunPython.noop
        ),
    ]



from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from utilities.choices import ChoiceSet


class BranchStatusChoices(ChoiceSet):
    NEW = 'new'
    PROVISIONING = 'provisioning'
    READY = 'ready'
    SYNCING = 'syncing'
    MIGRATING = 'migrating'
    MERGING = 'merging'
    REVERTING = 'reverting'
    MERGED = 'merged'
    ARCHIVED = 'archived'
    PENDING_MIGRATIONS = 'pending-migrations'
    FAILED = 'failed'

    CHOICES = (
        (NEW, _('New'), 'cyan'),
        (PROVISIONING, _('Provisioning'), 'orange'),
        (READY, _('Ready'), 'green'),
        (SYNCING, _('Syncing'), 'orange'),
        (MIGRATING, _('Migrating'), 'orange'),
        (MERGING, _('Merging'), 'orange'),
        (REVERTING, _('Reverting'), 'orange'),
        (MERGED, _('Merged'), 'blue'),
        (ARCHIVED, _('Archived'), 'gray'),
        (PENDING_MIGRATIONS, _('Pending Migrations'), 'red'),
        (FAILED, _('Failed'), 'red'),
    )

    TRANSITIONAL = (
        PROVISIONING,
        SYNCING,
        MIGRATING,
        MERGING,
        REVERTING,
    )

    WORKING = (
        NEW,
        READY,
        PENDING_MIGRATIONS,
        *TRANSITIONAL,
    )


class BranchMergeStrategyChoices(ChoiceSet):
    ITERATIVE = 'iterative'
    SQUASH = 'squash'

    CHOICES = (
        (ITERATIVE, _('Iterative')),
        (SQUASH, pgettext_lazy('The act of compressing multiple records into one', 'Squash')),
    )


class BranchEventTypeChoices(ChoiceSet):
    PROVISIONED = 'provisioned'
    SYNCED = 'synced'
    MIGRATED = 'migrated'
    MERGED = 'merged'
    REVERTED = 'reverted'
    ARCHIVED = 'archived'

    CHOICES = (
        (PROVISIONED, _('Provisioned'), 'green'),
        (SYNCED, _('Synced'), 'cyan'),
        (MIGRATED, _('Migrated'), 'purple'),
        (MERGED, _('Merged'), 'blue'),
        (REVERTED, _('Reverted'), 'orange'),
        (ARCHIVED, _('Archived'), 'gray'),
    )



import logging
from functools import partial

from core.choices import ObjectChangeActionChoices
from core.models import ObjectChange, ObjectType
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate, post_save, pre_delete
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from extras.events import process_event_rules
from extras.models import EventRule
from netbox.signals import post_clean
from utilities.exceptions import AbortRequest
from utilities.serialization import serialize_object

from netbox_branching import signals

from .choices import BranchStatusChoices
from .constants import INCLUDE_MODELS
from .contextvars import active_branch
from .events import *
from .models import Branch, ChangeDiff
from .utilities import deactivate_branch

__all__ = (
    'check_pending_migrations',
    'handle_branch_event',
    'record_change_diff',
    'validate_branch_deletion',
    'validate_branching_operations',
)


def check_object_accessible_in_branch(branch, model, object_id):
    """
    Check if an object is accessible for operations in a branch.

    An object is accessible if it either exists in main or was created in the branch.
    This prevents operations on objects that were deleted in main.

    Args:
        branch: The Branch instance
        model: The model class
        object_id: The primary key of the object

    Returns:
        True if the object is accessible (exists in main or was created in branch), False otherwise
    """
    # Exempt models listed in INCLUDE_MODELS. These are replicated when a new branch is provisioned but are not
    # managed directly.
    if f'{model._meta.app_label}.{model._meta.model_name}' in INCLUDE_MODELS:
        return True

    # Check whether the object exists in main
    with deactivate_branch():
        try:
            model.objects.get(pk=object_id)
        except model.DoesNotExist:
            pass
        else:
            return True

    # Object doesn't exist in main - check if it was created in the branch
    content_type = ContentType.objects.get_for_model(model)
    return ChangeDiff.objects.filter(
        branch=branch,
        object_type=content_type,
        object_id=object_id,
        action=ObjectChangeActionChoices.ACTION_CREATE
    ).exists()


@receiver(post_clean)
def validate_branching_operations(sender, instance, **kwargs):
    """
    Validate that branching operations are valid (e.g., not modifying deleted objects).
    """
    branch = active_branch.get()

    # Only validate if we're in a branch and this model supports branching
    if branch is None:
        return

    # Check if this model supports branching
    try:
        object_type = ObjectType.objects.get_for_model(instance.__class__)
        if 'branching' not in object_type.features:
            return
    except ObjectType.DoesNotExist:
        return

    # For updates, check if the object exists in main or was created in the branch
    if hasattr(instance, 'pk') and instance.pk is not None:
        model = instance.__class__
        if not check_object_accessible_in_branch(branch, model, instance.pk):
            # Object was deleted in main, not created in branch
            raise ValidationError(
                _(
                    "Cannot modify {model_name} '{object_name}' because it has been deleted in the main branch. "
                    "Sync with the main branch to update."
                ).format(
                    model_name=model._meta.verbose_name,
                    object_name=str(instance)
                )
            )


@receiver(post_save, sender=ObjectChange)
def record_change_diff(instance, **kwargs):
    """
    When an ObjectChange is created, create or update the relevant ChangeDiff for the active Branch.
    """
    logger = logging.getLogger('netbox_branching.signal_receivers.record_change_diff')

    branch = active_branch.get()
    content_type = instance.changed_object_type
    object_id = instance.changed_object_id

    # If this type of object does not support branching, return immediately.
    if 'branching' not in content_type.object_type.features:
        return

    # If this is a global change, update the "current" state in any ChangeDiffs for this object.
    if branch is None:

        # There cannot be a pre-existing ChangeDiff for an object that was just created.
        if instance.action == ObjectChangeActionChoices.ACTION_CREATE:
            return

        logger.debug(f"Updating change diff for global change to {instance.changed_object}")
        ChangeDiff.objects.filter(
            object_type=content_type,
            object_id=object_id,
            branch__status=BranchStatusChoices.READY
        ).update(
            last_updated=timezone.now(),
            current=instance.postchange_data_clean or None
        )

    # If this is a branch-aware change, create or update ChangeDiff for this object.
    else:

        # Updating the existing ChangeDiff
        if diff := ChangeDiff.objects.filter(object_type=content_type, object_id=object_id, branch=branch).first():
            logger.debug(f"Updating branch change diff for change to {instance.changed_object}")
            diff.object = instance.changed_object
            diff.last_updated = timezone.now()
            if diff.action != ObjectChangeActionChoices.ACTION_CREATE:
                diff.action = instance.action
            diff.modified = instance.postchange_data_clean or None
            diff.save()

        # Creating a new ChangeDiff
        else:
            logger.debug(f"Creating branch change diff for change to {instance.changed_object}")
            if instance.action == ObjectChangeActionChoices.ACTION_CREATE:
                current_data = None
            else:
                model = instance.changed_object_type.model_class()
                # For update operations, validate that object is accessible.
                if (
                    instance.action != ObjectChangeActionChoices.ACTION_DELETE and
                    not check_object_accessible_in_branch(branch, model, instance.changed_object_id)
                ):
                    # Object was deleted in main, not created in branch
                    raise AbortRequest(
                        _(
                            "Cannot {action} {model_name} '{object_name}' because it has been deleted "
                            "in the main branch. Sync with the main branch to update."
                        ).format(
                            action=instance.action.lower(),
                            model_name=model._meta.verbose_name,
                            object_name=str(instance.changed_object)
                        )
                    )

                # Check if object exists in main to determine if we need to get current_data
                with deactivate_branch():
                    try:
                        obj = model.objects.get(pk=instance.changed_object_id)
                        # Object exists in main, get its current state
                        if hasattr(obj, 'serialize_object'):
                            current_data = obj.serialize_object(exclude=['created', 'last_updated'])
                        else:
                            current_data = serialize_object(obj, exclude=['created', 'last_updated'])
                    except model.DoesNotExist:
                        # Object was created in branch, so there's no current state in main
                        current_data = None
            diff = ChangeDiff(
                branch=branch,
                object=instance.changed_object,
                action=instance.action,
                original=instance.prechange_data_clean or None,
                modified=instance.postchange_data_clean or None,
                current=current_data or None,
                last_updated=timezone.now(),
            )
            diff.save()


def handle_branch_event(event_type, branch, user=None, **kwargs):
    """
    Process any EventRules associated with branch events (e.g. syncing or merging).
    """
    logger = logging.getLogger('netbox_branching.signal_receivers.handle_branch_event')
    logger.debug(f"Checking for {event_type} event rules")

    # Find any EventRules for this event type
    object_type = ObjectType.objects.get_by_natural_key('netbox_branching', 'branch')
    event_rules = EventRule.objects.filter(
        event_types__contains=[event_type],
        enabled=True,
        object_types=object_type
    )
    if not event_rules:
        logger.debug("No matching event rules found")
        return
    logger.debug(f"Found {len(event_rules)} event rules")

    # Serialize the branch & process EventRules
    username = user.username if user else None
    data = serialize_object(branch)
    data['id'] = branch.pk
    process_event_rules(
        event_rules=event_rules,
        object_type=object_type,
        event_type=event_type,
        data=data,
        username=username
    )


# Connect signals with weak=False to prevent garbage collection of partial objects
signals.post_provision.connect(partial(handle_branch_event, event_type=BRANCH_PROVISIONED), weak=False)
signals.post_deprovision.connect(partial(handle_branch_event, event_type=BRANCH_DEPROVISIONED), weak=False)
signals.post_sync.connect(partial(handle_branch_event, event_type=BRANCH_SYNCED), weak=False)
signals.post_merge.connect(partial(handle_branch_event, event_type=BRANCH_MERGED), weak=False)
signals.post_revert.connect(partial(handle_branch_event, event_type=BRANCH_REVERTED), weak=False)


@receiver(pre_delete, sender=Branch)
def validate_branch_deletion(sender, instance, **kwargs):
    """
    Prevent the deletion of a Branch which is in a transitional state (e.g. provisioning, syncing, etc.).
    """
    if instance.status in BranchStatusChoices.TRANSITIONAL:
        raise AbortRequest(
            _("A branch in the {status} status may not be deleted.").format(status=instance.status)
        )


@receiver(post_migrate)
def check_pending_migrations(sender, using, **kwargs):
    """
    Check for any Branches with pending database migrations, and update their status accordingly.
    """
    if sender.name != 'netbox_branching' or using != DEFAULT_DB_ALIAS:
        return
    logger = logging.getLogger('netbox_branching.signal_receivers.check_pending_migrations')
    logger.info("Checking for branches with pending database migrations")

    open_branches = Branch.objects.filter(status=BranchStatusChoices.READY)
    update_count = 0
    for branch in open_branches:
        if branch.pending_migrations:
            branch.status = BranchStatusChoices.PENDING_MIGRATIONS
            update_count += 1
    if update_count:
        logger.info(f"Updating status of {update_count} branches with pending migrations")
        Branch.objects.bulk_update(open_branches, ['status'], batch_size=100)



from django.db import connection, migrations
from netbox.plugins import get_plugin_config

from netbox_branching.choices import BranchStatusChoices
from netbox_branching.utilities import get_sql_results

# Indexes to ignore as they are removed in a NetBox v4.3 migration
SKIP = (
    'dcim_cabletermination_termination_type_id_termination_id_idx',     # Removed in dcim.0207_remove_redundant_indexes
    'vpn_l2vpntermination_assigned_object_type_id_assigned_objec_idx',  # Removed in vpn.0009_remove_redundant_indexes
    'vpn_tunneltermination_termination_type_id_termination_id_idx',     # Removed in vpn.0009_remove_redundant_indexes
)


def rename_indexes(apps, schema_editor):
    """
    Rename all indexes within each branch to match the main schema.
    """
    Branch = apps.get_model('netbox_branching', 'Branch')
    schema_prefix = get_plugin_config('netbox_branching', 'schema_prefix')
    main_schema = get_plugin_config('netbox_branching', 'main_schema')

    with connection.cursor() as cursor:

        for branch in Branch.objects.filter(status=BranchStatusChoices.READY):
            print(f'\n    Renaming indexes for branch {branch.name} ({branch.schema_id})...', end='')
            schema_name = f'{schema_prefix}{branch.schema_id}'

            # Fetch all SQL indexes from the branch schema
            cursor.execute(
                f"SELECT tablename, indexname, indexdef FROM pg_indexes WHERE schemaname = '{schema_name}'"
            )
            branch_indexes = get_sql_results(cursor)

            for branch_index in branch_indexes:

                # Skip index if applicable
                if branch_index.indexname in SKIP:
                    continue

                # Find the matching index in main based on its table & definition
                definition = branch_index.indexdef.split(' USING ', maxsplit=1)[1]
                cursor.execute(
                    "SELECT indexname FROM pg_indexes WHERE schemaname=%s AND tablename=%s AND indexdef LIKE %s",
                    [main_schema, branch_index.tablename, f'% {definition}']
                )
                if result := cursor.fetchone():
                    new_name = result[0]
                    if new_name != branch_index.indexname:
                        sql = f"ALTER INDEX {schema_name}.{branch_index.indexname} RENAME TO {new_name}"
                        try:
                            cursor.execute(sql)
                        except Exception:
                            print(sql)
                            raise

    print('\n ', end='')  # Padding for final "OK"


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_branching', '0002_branch_schema_id_unique'),
    ]

    operations = [
        migrations.RunPython(
            code=rename_indexes,
            reverse_code=migrations.RunPython.noop
        ),
    ]



from django.urls import include, path
from utilities.urls import get_model_urls

from . import views

urlpatterns = [
    # Branches
    path('branches/', views.BranchListView.as_view(), name='branch_list'),
    path('branches/add/', views.BranchEditView.as_view(), name='branch_add'),
    path('branches/import/', views.BranchBulkImportView.as_view(), name='branch_bulk_import'),
    path('branches/edit/', views.BranchBulkEditView.as_view(), name='branch_bulk_edit'),
    path('branches/delete/', views.BranchBulkDeleteView.as_view(), name='branch_bulk_delete'),
    path('branches/<int:pk>/', include(get_model_urls('netbox_branching', 'branch'))),

    # Change diffs
    path('changes/', views.ChangeDiffListView.as_view(), name='changediff_list'),
]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("netbox_branching", "0006_tag_object_types"),
    ]

    operations = [
        migrations.AddField(
            model_name="branch",
            name="merge_strategy",
            field=models.CharField(blank=True, null=True, default=None, max_length=50),
        ),
    ]



import logging
from functools import cached_property

from core.choices import ObjectChangeActionChoices
from core.models import ObjectChange as ObjectChange_
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.postgres.fields import ArrayField
from django.db import DEFAULT_DB_ALIAS, models
from django.utils.translation import gettext_lazy as _
from utilities.querysets import RestrictedQuerySet
from utilities.serialization import deserialize_object

from netbox_branching.utilities import update_object

__all__ = (
    'AppliedChange',
    'ChangeDiff',
    'ObjectChange',
)


class ObjectChange(ObjectChange_):
    """
    Proxy model for NetBox's ObjectChange.
    """
    class Meta:
        proxy = True

    def migrate(self, branch, revert=False):
        """
        Run all applicable change data migrators for the given Branch, in sequence. This ensures that changes made prior
        to the application of a database migration can be successfully applied to (or reverted from) the main schema.
        """
        object_type = '.'.join(self.changed_object_type.natural_key())
        for migrator in branch.migrators.get(object_type, []):
            migrator(self, revert)

    def apply(self, branch, using=DEFAULT_DB_ALIAS, logger=None, skip_missing=False):
        """
        Apply the change using the specified database connection.
        """
        logger = logger or logging.getLogger('netbox_branching.models.ObjectChange.apply')
        model = self.changed_object_type.model_class()
        logger.info(f'Applying change {self.pk} using {using} ({self})')

        # Run data migrators
        self.migrate(branch)

        # Creating a new object
        if self.action == ObjectChangeActionChoices.ACTION_CREATE:
            if hasattr(model, 'deserialize_object'):
                instance = model.deserialize_object(self.postchange_data, pk=self.changed_object_id)
            else:
                instance = deserialize_object(model, self.postchange_data, pk=self.changed_object_id)
            try:
                instance.object.full_clean()
            except (FileNotFoundError) as e:
                # If a file was deleted later in this branch it will fail here
                # so we need to ignore it. We can assume the NetBox state is valid.
                logger.warning(f'Ignoring missing file: {e}')
            instance.save(using=using)

        # Modifying an object
        elif self.action == ObjectChangeActionChoices.ACTION_UPDATE:
            try:
                instance = model.objects.using(using).get(pk=self.changed_object_id)
                logger.debug(f'Updating {model._meta.verbose_name} {instance}')
                update_object(instance, self.diff()['post'], using=using)
            except model.DoesNotExist:
                if skip_missing:
                    logger.debug(f'{model._meta.verbose_name} ID {self.changed_object_id} already deleted; skipping')
                else:
                    raise

        # Deleting an object
        elif self.action == ObjectChangeActionChoices.ACTION_DELETE:
            try:
                instance = model.objects.get(pk=self.changed_object_id)
                logger.debug(f'Deleting {model._meta.verbose_name} {instance}')
                instance.delete(using=using)
            except model.DoesNotExist:
                logger.debug(f'{model._meta.verbose_name} ID {self.changed_object_id} already deleted; skipping')

    apply.alters_data = True

    def undo(self, branch, using=DEFAULT_DB_ALIAS, logger=None):
        """
        Revert a previously applied change using the specified database connection.
        """
        logger = logger or logging.getLogger('netbox_branching.models.ObjectChange.undo')
        model = self.changed_object_type.model_class()
        logger.info(f'Undoing change {self.pk} using {using} ({self})')

        # Run data migrators
        self.migrate(branch, revert=True)

        # Deleting a previously created object
        if self.action == ObjectChangeActionChoices.ACTION_CREATE:
            try:
                instance = model.objects.get(pk=self.changed_object_id)
                logger.debug(f'Undoing creation of {model._meta.verbose_name} {instance}')
                instance.delete(using=using)
            except model.DoesNotExist:
                logger.debug(f'{model._meta.verbose_name} ID {self.changed_object_id} does not exist; skipping')

        # Reverting a modification to an object
        elif self.action == ObjectChangeActionChoices.ACTION_UPDATE:
            instance = model.objects.using(using).get(pk=self.changed_object_id)
            update_object(instance, self.diff()['pre'], using=using)

        # Restoring a deleted object
        elif self.action == ObjectChangeActionChoices.ACTION_DELETE:
            deserialized = deserialize_object(model, self.prechange_data_clean, pk=self.changed_object_id)
            instance = deserialized.object
            logger.debug(f'Restoring {model._meta.verbose_name} {instance}')

            # Restore GenericForeignKey fields
            for field in instance._meta.private_fields:
                if isinstance(field, GenericForeignKey):
                    ct_field = getattr(instance, field.ct_field)
                    fk_field = getattr(instance, field.fk_field)
                    if ct_field and fk_field:
                        setattr(instance, field.name, ct_field.get_object_for_this_type(pk=fk_field))

            instance.full_clean()
            instance.save(using=using)

    undo.alters_data = True


class ChangeDiff(models.Model):
    branch = models.ForeignKey(
        to='netbox_branching.Branch',
        on_delete=models.CASCADE
    )
    last_updated = models.DateTimeField(
        auto_now_add=True
    )
    object_type = models.ForeignKey(
        to='contenttypes.ContentType',
        on_delete=models.PROTECT,
        related_name='+'
    )
    object_id = models.PositiveBigIntegerField()
    object = GenericForeignKey(
        ct_field='object_type',
        fk_field='object_id'
    )
    object_repr = models.CharField(
        max_length=200,
        editable=False
    )
    action = models.CharField(
        verbose_name=_('action'),
        max_length=50,
        choices=ObjectChangeActionChoices
    )
    original = models.JSONField(
        blank=True,
        null=True
    )
    modified = models.JSONField(
        blank=True,
        null=True
    )
    current = models.JSONField(
        blank=True,
        null=True
    )
    conflicts = ArrayField(
        base_field=models.CharField(max_length=100),
        editable=False,
        blank=True,
        null=True
    )

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ('-last_updated',)
        indexes = (
            models.Index(fields=('object_type', 'object_id')),
        )
        verbose_name = _('change diff')
        verbose_name_plural = _('change diffs')

    def __str__(self):
        return f'{self.get_action_display()} {self.object_type.name} {self.object_repr} ({self.object_id})'

    def save(self, *args, **kwargs):
        self._update_conflicts()
        if self.object:
            self.object_repr = str(self.object)

        super().save(*args, **kwargs)

    def get_action_color(self):
        return ObjectChangeActionChoices.colors.get(self.action)

    def _update_conflicts(self):
        """
        Record any conflicting changes between the modified and current object data.
        """
        if self.original is None or self.current is None:
            # Both the original and current states must be available to compare
            return
        conflicts = None
        if self.action == ObjectChangeActionChoices.ACTION_UPDATE:
            conflicts = [
                k for k, v in self.original.items()
                if v != self.modified[k] and v != self.current.get(k) and self.modified[k] != self.current.get(k)
            ]
        elif self.action == ObjectChangeActionChoices.ACTION_DELETE:
            conflicts = [
                k for k, v in self.original.items()
                if v != self.current.get(k)
            ]
        self.conflicts = conflicts or None

    @cached_property
    def altered_in_modified(self):
        """
        Return the set of attributes altered in the branch schema.
        """
        return {
            k for k, v in self.modified.items()
            if k in self.original and v != self.original[k]
        }

    @cached_property
    def altered_in_current(self):
        """
        Return the set of attributes altered in the main schema.
        """
        return {
            k for k, v in self.current.items()
            if k in self.original and v != self.original[k]
        }

    @cached_property
    def altered_fields(self):
        """
        Return an ordered list of attributes which have been modified in either the branch or main schema.
        """
        return sorted([*self.altered_in_modified, *self.altered_in_current])

    @cached_property
    def diff(self):
        """
        Provides a three-way summary of modified data, comparing the original, modified (branch), and current states.
        """
        return {
            'original': self.original_diff,
            'modified': self.modified_diff,
            'current': self.current_diff,
        }

    @cached_property
    def original_diff(self):
        """
        Return a key-value mapping of all attributes in the original state which have been modified.
        """
        return {
            k: v for k, v in self.original.items()
            if k in self.altered_fields
        }

    @cached_property
    def modified_diff(self):
        """
        Return a key-value mapping of all attributes which have been modified within the branch.
        """
        return {
            k: v for k, v in self.modified.items()
            if k in self.altered_fields
        }

    @cached_property
    def current_diff(self):
        """
        Return a key-value mapping of all attributes which have been modified outside the branch.
        """
        return {
            k: v for k, v in self.current.items()
            if k in self.altered_fields
        }


class AppliedChange(models.Model):
    """
    Maps an applied ObjectChange to a Branch.
    """
    change = models.OneToOneField(
        to='core.ObjectChange',
        on_delete=models.CASCADE,
        related_name='application'
    )
    branch = models.ForeignKey(
        to='netbox_branching.Branch',
        on_delete=models.CASCADE,
        related_name='applied_changes'
    )

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ('branch', 'change')
        verbose_name = _('applied change')
        verbose_name_plural = _('applied changes')



import datetime
import logging
from collections import namedtuple
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass
from functools import cached_property

from asgiref.local import Local
from django.contrib import messages
from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist
from django.db import connections
from django.db.models import ForeignKey, ManyToManyField
from django.http import HttpResponseBadRequest
from django.urls import reverse
from netbox.plugins import get_plugin_config
from netbox.utils import register_request_processor

from .constants import BRANCH_HEADER, COOKIE_NAME, EXEMPT_MODELS, EXEMPT_PATHS, INCLUDE_MODELS, QUERY_PARAM
from .contextvars import active_branch

# Thread-local storage for tracking branch connection aliases (matches Django's approach)
# Note: Aliases are tracked once and never removed, matching Django's pattern where
# DATABASES.keys() is static. Memory overhead is negligible (string references only).
_branch_connections_tracker = Local(thread_critical=False)

__all__ = (
    'ActiveBranchContextManager',
    'BranchActionIndicator',
    'ChangeSummary',
    'DynamicSchemaDict',
    'ListHandler',
    'activate_branch',
    'close_old_branch_connections',
    'deactivate_branch',
    'get_active_branch',
    'get_branchable_object_types',
    'get_sql_results',
    'get_tables_to_replicate',
    'is_api_request',
    'record_applied_change',
    'supports_branching',
    'track_branch_connection',
    'update_object',
)


def _get_tracked_branch_aliases():
    """Get set of tracked branch aliases for current thread."""
    if not hasattr(_branch_connections_tracker, 'aliases'):
        _branch_connections_tracker.aliases = set()
    return _branch_connections_tracker.aliases


def track_branch_connection(alias):
    """Register a branch connection alias for cleanup tracking."""
    _get_tracked_branch_aliases().add(alias)


class DynamicSchemaDict(dict):
    """
    Behaves like a normal dictionary, except for keys beginning with "schema_". Any lookup for
    "schema_*" will return the default configuration extended to include the search_path option.
    """
    @cached_property
    def main_schema(self):
        return get_plugin_config('netbox_branching', 'main_schema')

    def __getitem__(self, item):
        if type(item) is str and item.startswith('schema_') and (schema := item.removeprefix('schema_')):
            track_branch_connection(item)

            default_config = super().__getitem__('default')
            return {
                **default_config,
                'OPTIONS': {
                    **default_config.get('OPTIONS', {}),
                    'options': f'-c search_path={schema},{self.main_schema}'
                },
            }
        return super().__getitem__(item)

    def __contains__(self, item):
        if type(item) is str and item.startswith('schema_'):
            return True
        return super().__contains__(item)


def close_old_branch_connections(**kwargs):
    """
    Close branch database connections that have exceeded their maximum age.

    This function complements Django's close_old_connections() by handling
    dynamically-created branch connections. It tracks branch connection aliases
    in thread-local storage and closes them when they exceed CONN_MAX_AGE.

    Django's close_old_connections() only closes connections for database aliases
    found in DATABASES.keys(). Since branch aliases are generated dynamically and
    not present in that iteration (to avoid test isolation issues), they would never
    be cleaned up, causing connection leaks.

    This function is connected to request_started and request_finished signals,
    matching Django's cleanup timing.
    """

    for alias in _get_tracked_branch_aliases():
        conn = connections[alias]
        conn.close_if_unusable_or_obsolete()


@contextmanager
def activate_branch(branch):
    """
    A context manager for activating a Branch.
    """
    token = active_branch.set(branch)

    yield

    active_branch.reset(token)


@contextmanager
def deactivate_branch():
    """
    A context manager for temporarily deactivating the active Branch (if any). This is a
    convenience function for `activate_branch(None)`.
    """
    token = active_branch.set(None)

    yield

    active_branch.reset(token)


def get_branchable_object_types():
    """
    Return all object types which are branch-aware; i.e. those which support change logging.
    """
    from core.models import ObjectType

    return ObjectType.objects.with_feature('branching')


def supports_branching(model):
    """
    Returns True if branching is supported for the given model; otherwise False.
    """
    from netbox.models.features import ChangeLoggingMixin

    label = f'{model._meta.app_label}.{model._meta.model_name}'
    wildcard_label = f'{model._meta.app_label}.*'

    # Check for explicitly supported models
    if label in INCLUDE_MODELS:
        return True

    # Exclude models which do not support change logging
    if not issubclass(model, ChangeLoggingMixin):
        return False

    # TODO: Make this more efficient
    # Check for exempted models
    exempt_models = [
        *EXEMPT_MODELS,
        *get_plugin_config('netbox_branching', 'exempt_models', []),
    ]
    return label not in exempt_models and wildcard_label not in exempt_models


def get_tables_to_replicate():
    """
    Return an ordered list of database tables to replicate when provisioning a new schema.
    """
    tables = set()

    branch_aware_models = [
        ot.model_class() for ot in get_branchable_object_types()
        if ot.model_class() is not None
    ]
    for model in branch_aware_models:

        # Capture the model's table
        tables.add(model._meta.db_table)

        # Capture any M2M fields which reference other replicated models
        for m2m_field in model._meta.local_many_to_many:
            if hasattr(m2m_field, 'through'):
                # Field is actually a manager
                m2m_table = m2m_field.through._meta.db_table
            else:
                m2m_table = m2m_field._get_m2m_db_table(model._meta)
            tables.add(m2m_table)

    return sorted(tables)


class ListHandler(logging.Handler):
    """
    A logging handler which appends log messages to list passed on initialization.
    """
    def __init__(self, *args, queue, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = queue

    def emit(self, record):
        self.queue.append(self.format(record))


@dataclass
class ChangeSummary:
    """
    A record indicating the number of changes which were made between a start and end time.
    """
    start: datetime.datetime
    end: datetime.datetime
    count: int


def update_object(instance, data, using):
    """
    Set an attribute on an object depending on the type of model field.
    """
    # Avoid AppRegistryNotReady exception
    from taggit.managers import TaggableManager
    logger = logging.getLogger('netbox_branching.utilities.update_object')
    instance.snapshot()
    m2m_assignments = {}

    for attr, value in data.items():
        # Account for custom field data
        if attr == 'custom_fields':
            attr = 'custom_field_data'

        try:
            model_field = instance._meta.get_field(attr)
            field_cls = model_field.__class__
        except FieldDoesNotExist:
            field_cls = None

        if field_cls and issubclass(field_cls, ForeignKey):
            # Direct value assignment for ForeignKeys must be done by the field's concrete name
            setattr(instance, f'{attr}_id', value)
        elif field_cls and issubclass(field_cls, (ManyToManyField, TaggableManager)):
            # Use M2M manager for ManyToMany assignments
            m2m_manager = getattr(instance, attr)
            m2m_assignments[m2m_manager] = value
        else:
            setattr(instance, attr, value)

    try:
        instance.full_clean()
    except (FileNotFoundError) as e:
        # If a file was deleted later in this branch it will fail here
        # so we need to ignore it. We can assume the NetBox state is valid.
        logger.warning(f'Ignoring missing file: {e}')
    instance.save(using=using)

    for m2m_manager, value in m2m_assignments.items():
        m2m_manager.set(value)


def record_applied_change(instance, branch, **kwargs):
    """
    Create a new AppliedChange instance mapping an applied ObjectChange to its Branch.
    """
    from .models import AppliedChange

    AppliedChange.objects.update_or_create(change=instance, defaults={'branch': branch})


def is_api_request(request):
    """
    Returns True if the given request is a REST or GraphQL API request.
    """
    if not hasattr(request, 'path_info'):
        return False

    return request.path_info.startswith(reverse('api-root')) or request.path_info.startswith(reverse('graphql'))


def get_active_branch(request):
    """
    Return the active Branch (if any).
    """
    # The active Branch may be specified by HTTP header for REST & GraphQL API requests.
    from .models import Branch
    if is_api_request(request) and BRANCH_HEADER in request.headers:
        branch = Branch.objects.get(schema_id=request.headers.get(BRANCH_HEADER))
        if not branch.ready:
            return HttpResponseBadRequest(f"Branch {branch} is not ready for use (status: {branch.status})")
        return branch

    # Branch activated/deactivated by URL query parameter
    if QUERY_PARAM in request.GET:
        if schema_id := request.GET.get(QUERY_PARAM):
            branch = Branch.objects.get(schema_id=schema_id)
            if branch.ready:
                messages.success(request, f"Activated branch {branch}")
                return branch
            messages.error(request, f"Branch {branch} is not ready for use (status: {branch.status})")
            return None
        messages.success(request, "Deactivated branch")
        request.COOKIES.pop(COOKIE_NAME, None)  # Delete cookie if set
        return None

    # Branch set by cookie
    if schema_id := request.COOKIES.get(COOKIE_NAME):
        try:
            branch = Branch.objects.get(schema_id=schema_id)
            if branch.ready:
                return branch
        except ObjectDoesNotExist:
            pass
    return None


def get_sql_results(cursor):
    """
    Return the results of the most recent SQL query as a list of named tuples.
    """
    Result = namedtuple("Result", [col[0] for col in cursor.description])
    return [
        Result(*row) for row in cursor.fetchall()
    ]


@register_request_processor
def ActiveBranchContextManager(request):
    """
    Activate a branch if indicated by the request (except for exempt paths).
    """
    if request and request.path not in EXEMPT_PATHS and (branch := get_active_branch(request)):
        return activate_branch(branch)
    return nullcontext()


@dataclass
class BranchActionIndicator:
    """
    An indication of whether a particular branch action is permitted. If not, an explanatory message must be provided.
    """
    permitted: bool
    message: str = ''

    def __bool__(self):
        return self.permitted



from core.api.serializers import JobSerializer
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseBadRequest
from drf_spectacular.utils import extend_schema
from netbox.api.viewsets import BaseViewSet, NetBoxReadOnlyModelViewSet
from netbox.plugins import get_plugin_config
from rest_framework.decorators import action
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.response import Response
from rest_framework.routers import APIRootView
from rest_framework.viewsets import ModelViewSet

from netbox_branching import filtersets
from netbox_branching.jobs import MergeBranchJob, RevertBranchJob, SyncBranchJob
from netbox_branching.models import Branch, BranchEvent, ChangeDiff

from . import serializers


class RootView(APIRootView):
    def get_view_name(self):
        return 'Branching'


class BranchViewSet(ModelViewSet):
    queryset = Branch.objects.all()
    serializer_class = serializers.BranchSerializer
    filterset_class = filtersets.BranchFilterSet

    @extend_schema(
        methods=['post'],
        request=serializers.CommitSerializer(),
        responses={200: JobSerializer()},
    )
    @action(detail=True, methods=['post'])
    def sync(self, request, pk):
        """
        Enqueue a background job to synchronize a branch from main.
        """
        if not request.user.has_perm('netbox_branching.sync_branch'):
            raise PermissionDenied("This user does not have permission to sync branches.")

        branch = self.get_object()
        if not branch.ready:
            return HttpResponseBadRequest("Branch is not ready to sync.")

        serializer = serializers.CommitSerializer(data=request.data)
        commit = serializer.validated_data.get('commit', True) if serializer.is_valid() else False

        # Enqueue a background job
        job = SyncBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=commit
        )

        return Response(JobSerializer(job, context={'request': request}).data)

    @extend_schema(
        methods=['post'],
        request=serializers.CommitSerializer(),
        responses={200: JobSerializer()},
    )
    @action(detail=True, methods=['post'])
    def merge(self, request, pk):
        """
        Enqueue a background job to merge a branch.
        """
        if not request.user.has_perm('netbox_branching.merge_branch'):
            raise PermissionDenied("This user does not have permission to merge branches.")

        branch = self.get_object()
        if not branch.ready:
            return HttpResponseBadRequest("Branch is not ready to merge.")

        serializer = serializers.CommitSerializer(data=request.data)
        commit = serializer.validated_data.get('commit', True) if serializer.is_valid() else False

        # Enqueue a background job
        job = MergeBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=commit,
            job_timeout=get_plugin_config('netbox_branching', 'job_timeout')
        )

        return Response(JobSerializer(job, context={'request': request}).data)

    @extend_schema(
        methods=['post'],
        request=serializers.CommitSerializer(),
        responses={200: JobSerializer()},
    )
    @action(detail=True, methods=['post'])
    def revert(self, request, pk):
        """
        Enqueue a background job to revert a merged branch.
        """
        if not request.user.has_perm('netbox_branching.revert_branch'):
            raise PermissionDenied("This user does not have permission to revert branches.")

        branch = self.get_object()
        if not branch.merged:
            return HttpResponseBadRequest("Only merged branches can be reverted.")

        serializer = serializers.CommitSerializer(data=request.data)
        commit = serializer.validated_data.get('commit', True) if serializer.is_valid() else False

        # Enqueue a background job
        job = RevertBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=commit
        )

        return Response(JobSerializer(job, context={'request': request}).data)

    @extend_schema(
        methods=['post'],
        responses={200: serializers.BranchSerializer()},
    )
    @action(detail=True, methods=['post'])
    def archive(self, request, pk):
        """
        Archive a merged branch, deprovisioning its schema.
        """
        if not request.user.has_perm('netbox_branching.archive_branch'):
            raise PermissionDenied("This user does not have permission to archive branches.")

        branch = self.get_object()
        if not branch.merged:
            return HttpResponseBadRequest("Only merged branches can be archived.")
        if not branch.can_archive:
            return HttpResponseBadRequest("Archiving this branch is not permitted.")

        branch.archive(user=request.user)
        branch.refresh_from_db()

        serializer = self.get_serializer(branch)
        return Response(serializer.data)


class BranchEventViewSet(ListModelMixin, RetrieveModelMixin, BaseViewSet):
    queryset = BranchEvent.objects.all()
    serializer_class = serializers.BranchEventSerializer
    filterset_class = filtersets.BranchEventFilterSet


class ChangeDiffViewSet(NetBoxReadOnlyModelViewSet):
    queryset = ChangeDiff.objects.all()
    serializer_class = serializers.ChangeDiffSerializer
    filterset_class = filtersets.ChangeDiffFilterSet



from django.utils.translation import gettext as _
from netbox.events import EVENT_TYPE_KIND_SUCCESS, EventType

__all__ = (
    'BRANCH_DEPROVISIONED',
    'BRANCH_MERGED',
    'BRANCH_PROVISIONED',
    'BRANCH_REVERTED',
    'BRANCH_SYNCED',
)

# Branch events
BRANCH_PROVISIONED = 'branch_provisioned'
BRANCH_DEPROVISIONED = 'branch_deprovisioned'
BRANCH_SYNCED = 'branch_synced'
BRANCH_MERGED = 'branch_merged'
BRANCH_REVERTED = 'branch_reverted'


# Register core events
EventType(BRANCH_PROVISIONED, _('Branch provisioned')).register()
EventType(BRANCH_DEPROVISIONED, _('Branch deprovisioned')).register()
EventType(BRANCH_SYNCED, _('Branch synced'), kind=EVENT_TYPE_KIND_SUCCESS).register()
EventType(BRANCH_MERGED, _('Branch merged'), kind=EVENT_TYPE_KIND_SUCCESS).register()
EventType(BRANCH_REVERTED, _('Branch reverted'), kind=EVENT_TYPE_KIND_SUCCESS).register()



from django.dispatch import Signal

__all__ = (
    'post_deprovision',
    'post_merge',
    'post_migrate',
    'post_provision',
    'post_revert',
    'post_sync',
    'pre_deprovision',
    'pre_merge',
    'pre_migrate',
    'pre_provision',
    'pre_revert',
    'pre_sync',
)

# Pre-event signals
pre_provision = Signal()
pre_deprovision = Signal()
pre_sync = Signal()
pre_migrate = Signal()
pre_merge = Signal()
pre_revert = Signal()

# Post-event signals
post_provision = Signal()
post_deprovision = Signal()
post_sync = Signal()
post_migrate = Signal()
post_merge = Signal()
post_revert = Signal()



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_branching', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='branch',
            name='schema_id',
            field=models.CharField(editable=False, max_length=8, unique=True),
        ),
    ]



from django.utils.translation import gettext_lazy as _
from netbox.plugins import PluginMenu, PluginMenuButton, PluginMenuItem

menu = PluginMenu(
    label='Branching',
    groups=(
        (_('Branches'), (
            PluginMenuItem(
                link='plugins:netbox_branching:branch_list',
                link_text=_('Branches'),
                auth_required=True,
                buttons=(
                    PluginMenuButton('plugins:netbox_branching:branch_add', _('Add'), 'mdi mdi-plus-thick'),
                    PluginMenuButton('plugins:netbox_branching:branch_bulk_import', _('Import'), 'mdi mdi-upload'),
                )
            ),
            PluginMenuItem(
                link='plugins:netbox_branching:changediff_list',
                link_text=_('Change Diffs'),
                auth_required=True
            ),
        )),
    ),
    icon_class='mdi mdi-source-branch'
)



import logging
from abc import ABC, abstractmethod

from mptt.models import MPTTModel

__all__ = (
    'MergeStrategy',
    'get_merge_strategy',
)


class MergeStrategy(ABC):
    """
    Abstract base class for merge strategies.
    """
    # Ordering for changes queryset in revert() operation
    # Merge always uses chronological order ('time')
    revert_changes_ordering = '-time'  # Reverse chronological order (newest first)

    @abstractmethod
    def merge(self, branch, changes, request, logger, user):
        """
        Merge changes from the branch into the main schema.

        Args:
            branch: The Branch instance being merged
            changes: QuerySet of ObjectChanges to merge
            request: Django request object for event tracking
            logger: Logger instance for logging
            user: User who initiated the merge
        """

    @abstractmethod
    def revert(self, branch, changes, request, logger, user):
        """
        Revert changes that were previously merged.

        Args:
            branch: The Branch instance being reverted
            changes: QuerySet of ObjectChanges to revert
            request: Django request object for event tracking
            logger: Logger instance for logging
            user: User who initiated the revert
        """

    def _clean(self, models):
        """
        Called after syncing, merging, or reverting a branch.
        """
        logger = logging.getLogger('netbox_branching.branch')

        for model in models:

            # Recalculate MPTT as needed
            if issubclass(model, MPTTModel):
                logger.debug(f"Recalculating MPTT for model {model}")
                model.objects.rebuild()


def get_merge_strategy(strategy_name):
    """
    Get the appropriate merge strategy class based on the strategy name.

    Args:
        strategy_name: String name of the strategy from BranchMergeStrategyChoices, or None

    Returns:
        MergeStrategy class (caller should instantiate)

    Raises:
        ValueError: If the strategy name is unknown
    """
    from netbox_branching.choices import BranchMergeStrategyChoices

    from .iterative import IterativeMergeStrategy
    from .squash import SquashMergeStrategy

    # Default to ITERATIVE if strategy_name is None
    if strategy_name is None:
        strategy_name = BranchMergeStrategyChoices.ITERATIVE

    strategies = {
        BranchMergeStrategyChoices.SQUASH: SquashMergeStrategy,
        BranchMergeStrategyChoices.ITERATIVE: IterativeMergeStrategy,
    }

    try:
        return strategies[strategy_name]
    except KeyError as exc:
        raise ValueError(f"Invalid strategy name: {strategy_name}") from exc



from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseBadRequest, HttpResponseRedirect

from .constants import COOKIE_NAME, EXEMPT_PATHS, QUERY_PARAM
from .utilities import get_active_branch, is_api_request

__all__ = (
    'BranchMiddleware',
)


class BranchMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def _apply_branch_cookie(self, response, branch, branch_change):
        """
        Apply or remove the branch cookie on the given response.
        """
        if branch:
            response.set_cookie(COOKIE_NAME, branch.schema_id)
        elif branch_change:
            response.delete_cookie(COOKIE_NAME)

    def __call__(self, request):

        # Skip branch activation for exempt paths
        if request.path in EXEMPT_PATHS:
            return self.get_response(request)

        # Set/clear the active Branch on the request
        try:
            branch = get_active_branch(request)
        except ObjectDoesNotExist:
            return HttpResponseBadRequest("Invalid branch identifier")

        response = self.get_response(request)

        # Set/clear the branch cookie (for non-API requests)
        if not is_api_request(request):
            # Check if a branch is being activated/deactivated
            branch_change = QUERY_PARAM in request.GET

            # Redirect to dashboard if branch activation/deactivation results in 404
            if branch_change and response.status_code == 404:
                # Construct a more informative error message
                branch_name = f"branch '{branch.name}'" if branch else "main"
                requested_url = request.path
                messages.warning(
                    request,
                    f"The requested object at {requested_url} does not exist in {branch_name}."
                )

                # Create redirect response and apply cookie operations to it
                redirect_response = HttpResponseRedirect('/')
                self._apply_branch_cookie(redirect_response, branch, branch_change)
                return redirect_response

            # Set/clear cookie on the normal response
            self._apply_branch_cookie(response, branch, branch_change)

        return response



import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_branching', '0004_copy_migrations'),
    ]

    operations = [
        migrations.AddField(
            model_name='branch',
            name='applied_migrations',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=200),
                blank=True,
                default=list,
                size=None
            ),
        ),
    ]



"""
Merge strategy implementations for branch operations.
"""
from .iterative import IterativeMergeStrategy
from .squash import SquashMergeStrategy
from .strategy import MergeStrategy, get_merge_strategy

__all__ = (
    'IterativeMergeStrategy',
    'MergeStrategy',
    'SquashMergeStrategy',
    'get_merge_strategy',
)



from django.urls import reverse_lazy

__all__ = (
    'BRANCH_ACTIONS',
    'BRANCH_HEADER',
    'COOKIE_NAME',
    'EXEMPT_MODELS',
    'EXEMPT_PATHS',
    'INCLUDE_MODELS',
    'QUERY_PARAM',
    'SKIP_INDEXES',
)


# HTTP cookie
COOKIE_NAME = 'active_branch'

# HTTP header for API requests
BRANCH_HEADER = 'X-NetBox-Branch'

# Branch actions
BRANCH_ACTIONS = (
    'sync',
    'merge',
    'migrate',
    'revert',
    'archive',
)

# Paths exempt from branch activation
EXEMPT_PATHS = (
    reverse_lazy('api-status'),
)

# URL query parameter name
QUERY_PARAM = '_branch'

# Models which do not support change logging, but whose database tables
# must be replicated for each branch to ensure proper functionality
INCLUDE_MODELS = (
    'dcim.cablepath',
    'extras.cachedvalue',
    'extras.taggeditem',  # Fix for issue #354 - tags through model
    'tenancy.contactgroupmembership',  # Fix for NetBox v4.3.0
)

# Models for which branching support is explicitly disabled
EXEMPT_MODELS = (
    # Exempt applicable core NetBox models
    'core.*',
    'extras.branch',
    'extras.customfield',
    'extras.customfieldchoiceset',
    'extras.customlink',
    'extras.eventrule',
    'extras.exporttemplate',
    'extras.notificationgroup',
    'extras.savedfilter',
    'extras.webhook',

    # Exempt all models from this plugin and from netbox-changes
    'netbox_branching.*',
    'netbox_changes.*',
)

# Indexes to ignore as they are removed in a NetBox v4.3 migration, but might be present
# in earlier NetBox releases.
# TODO: Remove in v0.6.0
SKIP_INDEXES = (
    'dcim_cabletermination_termination_type_id_termination_id_idx',     # Removed in dcim.0207_remove_redundant_indexes
    'vpn_l2vpntermination_assigned_object_type_id_assigned_objec_idx',  # Removed in vpn.0009_remove_redundant_indexes
    'vpn_tunneltermination_termination_type_id_termination_id_idx',     # Removed in vpn.0009_remove_redundant_indexes
)



import django_tables2 as tables
from core.tables import ObjectChangeTable
from django.utils.translation import gettext_lazy as _
from utilities.tables import register_table_column

__all__ = (
    'ConflictsColumn',
    'DiffColumn',
)


OBJECTCHANGE_FULL_NAME = """
{% load helpers %}
{{ value.get_full_name|placeholder }}
"""

OBJECTCHANGE_OBJECT = """
{% if value and value.get_absolute_url %}
    <a href="{{ value.get_absolute_url }}">{{ record.object_repr }}</a>
{% else %}
    {{ record.object_repr }}
{% endif %}
"""

BEFORE_DIFF = """
{% if record.action == 'create' %}
    {{ ''|placeholder }}
{% else %}
    <pre class="p-0">{% for k, v in record.diff.pre.items %}{{ k }}: {{ v }}
{% endfor %}</pre>
{% endif %}
"""

AFTER_DIFF = """
{% if record.action == 'delete' %}
    {{ ''|placeholder }}
{% else %}
    <pre class="p-0">{% for k, v in record.diff.post.items %}{{ k }}: {{ v }}
{% endfor %}</pre>
{% endif %}
"""


class ConflictsColumn(tables.TemplateColumn):
    template_code = """
    {% if record.conflicts %}
      <span class="text-red"><i class="mdi mdi-alert-octagon"></i></span>
    {% else %}
      {{ ''|placeholder }}
    {% endif %}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, template_code=self.template_code, **kwargs)


class DiffColumn(tables.TemplateColumn):
    template_code = """{% load helpers %}
        {% for k, v in value.items %}
        {{ k }}:
          {% if show_conflicts and k in record.conflicts %}
            <span class="bg-red text-red-fg px-1 rounded-2">{{ v|placeholder }}</span>
          {% elif v != record.original|get_item:k %}
            <span class="bg-green text-green-fg px-1 rounded-2">{{ v|placeholder }}</span>
          {% else %}
            {{ v|placeholder }}
          {% endif %}
        <br />{% endfor %}
        """

    def __init__(self, show_conflicts=True, *args, **kwargs):
        context = {
            'show_conflicts': show_conflicts,
        }
        super().__init__(*args, template_code=self.template_code, extra_context=context, **kwargs)

    def value(self, value):
        return str(value) if value else None


# Register a table column to include the Branch associated with each record in ObjectChangeTable
branch_column = tables.Column(
    accessor=tables.A('application__branch'),
    linkify=True,
    verbose_name=_('Branch')
)
register_table_column(branch_column, 'branch', ObjectChangeTable)



"""
Squash merge strategy implementation with functions for collapsing and ordering ObjectChanges.
"""
from enum import StrEnum

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, models
from netbox.context_managers import event_tracking

from .strategy import MergeStrategy

__all__ = (
    'SquashMergeStrategy',
)


class ActionType(StrEnum):
    """
    Enum for collapsed change action types.
    """
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'
    SKIP = 'skip'


class CollapsedChange:
    """
    Represents a collapsed set of ObjectChanges for a single object.
    """
    def __init__(self, key, model_class):
        self.key = key  # (content_type_id, object_id)
        self.model_class = model_class
        self.change_count = 0  # Number of changes processed
        self.final_action = None  # ActionType enum value or None
        self.prechange_data = {}
        self.postchange_data = {}
        self.last_change = None  # The most recent ObjectChange (for metadata)

        # Dependencies for ordering
        self.depends_on = set()  # Set of keys this change depends on
        self.depended_by = set()  # Set of keys that depend on this change

    def __repr__(self):
        obj_id = self.key[1]
        suffix = f" ({self.key[2]})" if len(self.key) > 2 else ""
        return (
            f"<CollapsedChange {self.model_class.__name__}:{obj_id}{suffix} "
            f"action={self.final_action} changes={self.change_count}>"
        )

    def add_change(self, change, logger):
        """
        Incrementally process a single ObjectChange and update the collapsed state.
        This processes changes one-by-one to avoid storing all ObjectChanges in memory.

        Assumes changes are added in chronological order.
        """
        self.change_count += 1
        self.last_change = change

        # Dispatch to specific handler based on action type
        if change.action == 'create':
            self._add_change_create(change, logger)
        elif change.action == 'update':
            self._add_change_update(change, logger)
        elif change.action == 'delete':
            self._add_change_delete(change, logger)

    def _add_change_create(self, change, logger):
        """Handle CREATE action."""
        if self.final_action is None:
            logger.debug(f"  [{self.change_count}] CREATE")
            self.final_action = ActionType.CREATE
            self._set_initial_data(change)
        else:
            logger.warning(f"  [{self.change_count}] Unexpected CREATE after {self.final_action} for {self}")

    def _add_change_update(self, change, logger):
        """Handle UPDATE action."""
        if self.final_action is None:
            self.final_action = ActionType.UPDATE
            self._set_initial_data(change)
        elif self.final_action in (ActionType.CREATE, ActionType.UPDATE):
            logger.debug(f"  [{self.change_count}] UPDATE after {self.final_action} (still {self.final_action})")
            if change.postchange_data:
                self.postchange_data.update(change.postchange_data)
        else:
            logger.warning(f"  [{self.change_count}] Unexpected UPDATE after {self.final_action} for {self}")

    def _add_change_delete(self, change, logger):
        """Handle DELETE action."""
        if self.final_action is None:
            self.final_action = ActionType.DELETE
            self._set_initial_data(change)
        elif self.final_action == ActionType.CREATE:
            # CREATE + DELETE = SKIP
            logger.debug(f"  [{self.change_count}] DELETE after CREATE -> SKIP")
            self.final_action = ActionType.SKIP
            self.prechange_data = {}
            self.postchange_data = {}
        elif self.final_action == ActionType.UPDATE:
            # UPDATE + DELETE = DELETE
            self.final_action = ActionType.DELETE
            self.postchange_data = change.postchange_data
        else:
            # DELETE after DELETE or SKIP is unexpected
            logger.warning(f"  [{self.change_count}] Unexpected DELETE after {self.final_action} for {self}")

    def _set_initial_data(self, change):
        """Helper to set initial pre/postchange data for first CREATE or UPDATE."""
        self.prechange_data = change.prechange_data or {}
        if change.postchange_data:
            self.postchange_data.update(change.postchange_data)

    def generate_object_change(self):
        """
        Generate a dummy ObjectChange instance from this collapsed change.
        Used to leverage the standard ObjectChange.apply() and undo() methods.
        """
        from netbox_branching.models import ObjectChange
        app_label, model = self.key[0].split('.')
        dummy_change = ObjectChange(
            action=self.final_action.value if self.final_action else None,
            changed_object_type=ContentType.objects.get_by_natural_key(app_label, model),
            changed_object_id=self.key[1],
            prechange_data=self.prechange_data,
            postchange_data=self.postchange_data,
        )
        # Use last_change for migrate() to have the correct metadata
        dummy_change.pk = self.last_change.pk
        return dummy_change


class SquashMergeStrategy(MergeStrategy):
    """
    Squash merge strategy that collapses multiple changes per object into a single operation.
    """
    # Override: squash strategy needs chronological order for both merge and revert
    # because the collapse logic expects CREATE -> UPDATE -> DELETE order
    revert_changes_ordering = 'time'

    @staticmethod
    def _collapse_changes(changes, logger):
        """
        Collapse a queryset of ObjectChanges into a dict of CollapsedChange objects keyed by
        (model_label, object_id). Returns a tuple of (collapsed_changes, change_count).
        """
        collapsed_changes = {}
        change_count = 0

        for change in changes:
            change_count += 1
            app_label, model = change.changed_object_type.natural_key()
            key = (f"{app_label}.{model}", change.changed_object_id)

            if key not in collapsed_changes:
                model_class = change.changed_object_type.model_class()
                collapsed_changes[key] = CollapsedChange(key, model_class)
                logger.debug(f"New object: {model_class.__name__}:{change.changed_object_id}")

            collapsed_changes[key].add_change(change, logger)

        return collapsed_changes, change_count

    @staticmethod
    def _skip_updates_missing_in_main(collapsed_changes, logger):
        """
        Mark any collapsed UPDATE as SKIP if the object no longer exists in main. This handles
        the case where an object was modified in the branch but deleted in main and then synced,
        leaving only an UPDATE in the branch's ObjectChange log with no object to act on.
        """
        for collapsed in collapsed_changes.values():
            if collapsed.final_action == ActionType.UPDATE:
                exists = collapsed.model_class.objects.using(DEFAULT_DB_ALIAS).filter(
                    pk=collapsed.key[1]
                ).exists()
                if not exists:
                    logger.info(
                        f"  Skipping UPDATE for {collapsed.model_class.__name__}:{collapsed.key[1]} "
                        f"(object deleted in main)"
                    )
                    collapsed.final_action = ActionType.SKIP

    def merge(self, branch, changes, request, logger, user):
        """
        Apply changes after collapsing them by object and ordering by dependencies.
        """
        models = set()

        logger.info("Collapsing ObjectChanges by object (incremental)...")
        collapsed_changes, _ = SquashMergeStrategy._collapse_changes(changes, logger)
        SquashMergeStrategy._skip_updates_missing_in_main(collapsed_changes, logger)

        # Order collapsed changes based on dependencies
        ordered_changes = SquashMergeStrategy._order_collapsed_changes(collapsed_changes, logger)

        # Apply collapsed changes in order
        logger.info(f"Applying {len(ordered_changes)} collapsed changes...")
        for i, collapsed in enumerate(ordered_changes, 1):
            model_class = collapsed.model_class
            models.add(model_class)

            last_change = collapsed.last_change

            logger.info(f"  [{i}/{len(ordered_changes)}] {collapsed.final_action.upper()} "
                       f"{model_class.__name__}:{collapsed.key[1]} "
                       f"(from {collapsed.change_count} original changes)")

            with event_tracking(request):
                request.id = last_change.request_id
                request.user = user

                # Create a dummy ObjectChange from the collapsed change and apply it
                dummy_change = collapsed.generate_object_change()
                dummy_change.apply(branch, using=DEFAULT_DB_ALIAS, logger=logger)

        # Perform cleanup tasks
        self._clean(models)

    def revert(self, branch, changes, request, logger, user):
        """
        Undo changes after collapsing them by object and ordering by dependencies.
        """
        models = set()

        logger.info("Collapsing ObjectChanges by object (incremental)...")
        collapsed_changes, change_count = SquashMergeStrategy._collapse_changes(changes, logger)
        logger.info(f"  {change_count} changes collapsed into {len(collapsed_changes)} objects")
        SquashMergeStrategy._skip_updates_missing_in_main(collapsed_changes, logger)

        # Order collapsed changes for revert (reverse of merge order)
        merge_order = SquashMergeStrategy._order_collapsed_changes(collapsed_changes, logger)
        ordered_changes = list(reversed(merge_order))

        # Undo collapsed changes in dependency order
        logger.info(f"Undoing {len(ordered_changes)} collapsed changes in dependency order...")
        for i, collapsed in enumerate(ordered_changes, 1):
            model_class = collapsed.model_class
            models.add(model_class)

            # Use the last change's metadata for tracking
            last_change = collapsed.last_change
            logger.info(
                f"[{i}/{len(ordered_changes)}] Undoing {collapsed.final_action} "
                f"{model_class._meta.verbose_name} (ID: {collapsed.key[1]})"
            )

            with event_tracking(request):
                request.id = last_change.request_id
                request.user = user

                # Create a dummy ObjectChange from the collapsed change and undo it
                dummy_change = collapsed.generate_object_change()
                dummy_change.undo(branch, using=DEFAULT_DB_ALIAS, logger=logger)

        # Perform cleanup tasks
        self._clean(models)

    @staticmethod
    def _get_fk_references(model_class, data, changed_objects):
        """
        Find foreign key references in the given data that point to objects in changed_objects.
        Returns a set of (model_label, object_id) tuples where model_label is "app.model".
        Handles both regular ForeignKey fields and GenericForeignKey fields.
        """
        if not data:
            return set()

        references = set()

        # Check regular ForeignKey fields
        for field in model_class._meta.get_fields():
            if isinstance(field, models.ForeignKey):
                fk_value = data.get(field.name)

                if fk_value:
                    # Get the content type of the related model
                    related_model = field.related_model
                    related_ct = ContentType.objects.get_for_model(related_model)
                    app_label, model = related_ct.natural_key()
                    model_label = f"{app_label}.{model}"
                    ref_key = (model_label, fk_value)

                    # Only track if this object is in our changed_objects
                    if ref_key in changed_objects:
                        references.add(ref_key)

        # Check GenericForeignKey fields
        for field in model_class._meta.private_fields:
            if isinstance(field, GenericForeignKey):
                # ObjectChange data may store the CT FK as either 'field_name' or 'field_name_id'
                ct_value = data.get(field.ct_field) or data.get(field.ct_field + '_id')
                fk_value = data.get(field.fk_field)

                if ct_value and fk_value:
                    try:
                        ct = ContentType.objects.get_for_id(ct_value)
                        app_label, model = ct.natural_key()
                        model_label = f"{app_label}.{model}"
                        ref_key = (model_label, fk_value)

                        if ref_key in changed_objects:
                            references.add(ref_key)
                    except ContentType.DoesNotExist:
                        pass

        return references

    @staticmethod
    def _build_fk_dependency_graph(deletes, updates, creates, logger):
        """
        Build the FK dependency graph between collapsed changes.

        Analyzes foreign key references in the data to determine which changes depend on others:
        - UPDATEs that remove FK references must happen before DELETEs
        - UPDATEs that add FK references must happen after CREATEs
        - CREATEs that reference other created objects must happen after those CREATEs
        - DELETEs of child objects must happen before DELETEs of parent objects

        Modifies the CollapsedChange objects in place by setting their depends_on and depended_by sets.
        """
        # Build lookup maps for efficient dependency checking
        deletes_map = {c.key: c for c in deletes}
        creates_map = {c.key: c for c in creates}

        # 1. Check UPDATEs for dependencies
        for update in updates:
            # Check if UPDATE references deleted object in prechange_data
            # This means the UPDATE had a reference that it's removing
            # The UPDATE must happen BEFORE the DELETE so the FK reference is removed first
            if update.prechange_data:
                prechange_refs = SquashMergeStrategy._get_fk_references(
                    update.model_class,
                    update.prechange_data,
                    deletes_map.keys()
                )
                for ref_key in prechange_refs:
                    # DELETE depends on UPDATE (UPDATE removes reference, then DELETE can proceed)
                    delete_collapsed = deletes_map[ref_key]
                    delete_collapsed.depends_on.add(update.key)
                    update.depended_by.add(ref_key)
                    logger.debug(
                        f"    {delete_collapsed} depends on {update} "
                        f"(UPDATE removes FK reference before DELETE)"
                    )

            # Check if UPDATE references created object in postchange_data
            # This means the UPDATE needs the CREATE to exist first
            # The CREATE must happen BEFORE the UPDATE
            if update.postchange_data:
                postchange_refs = SquashMergeStrategy._get_fk_references(
                    update.model_class,
                    update.postchange_data,
                    creates_map.keys()
                )
                for ref_key in postchange_refs:
                    # UPDATE depends on CREATE
                    create_collapsed = creates_map[ref_key]
                    update.depends_on.add(ref_key)
                    create_collapsed.depended_by.add(update.key)
                    logger.debug(
                        f"    {update} depends on {create_collapsed} "
                        f"(UPDATE references created object)"
                    )

        # 2. Check CREATEs for dependencies on other CREATEs
        for create in creates:
            if create.postchange_data:
                # Check if this CREATE references other created objects
                refs = SquashMergeStrategy._get_fk_references(
                    create.model_class,
                    create.postchange_data,
                    creates_map.keys()
                )
                for ref_key in refs:
                    if ref_key != create.key:  # Don't self-reference
                        # CREATE depends on another CREATE
                        ref_create = creates_map[ref_key]
                        create.depends_on.add(ref_key)
                        ref_create.depended_by.add(create.key)
                        logger.debug(
                            f"    {create} depends on {ref_create} "
                            f"(CREATE references another created object)"
                        )

        # 3. Check DELETEs for dependencies on other DELETEs
        for delete in deletes:
            if delete.prechange_data:
                # Check if this DELETE references other deleted objects
                refs = SquashMergeStrategy._get_fk_references(
                    delete.model_class,
                    delete.prechange_data,
                    deletes_map.keys()
                )
                for ref_key in refs:
                    if ref_key != delete.key:  # Don't self-reference
                        # This delete references another deleted object
                        # The referenced object must be deleted AFTER this one (child before parent)
                        ref_delete = deletes_map[ref_key]
                        ref_delete.depends_on.add(delete.key)
                        delete.depended_by.add(ref_key)
                        logger.debug(
                            f"    {ref_delete} depends on {delete} "
                            f"(child DELETE must happen before parent DELETE)"
                        )

    @staticmethod
    def _has_fk_to(collapsed, target_model_class, target_obj_id):
        """
        Check if a CollapsedChange has a foreign key reference to a specific object.
        Returns True if any FK or GenericFK field in postchange_data points to the target object.
        """
        if not collapsed.postchange_data:
            return False

        for field in collapsed.model_class._meta.get_fields():
            if isinstance(field, models.ForeignKey):
                fk_value = collapsed.postchange_data.get(field.name)
                if fk_value:
                    # Check if this FK points to target model
                    related_model = field.related_model
                    if related_model == target_model_class and fk_value == target_obj_id:
                        return True

        for field in collapsed.model_class._meta.private_fields:
            if isinstance(field, GenericForeignKey):
                # ObjectChange data may store the CT FK as either 'field_name' or 'field_name_id'
                ct_value = (collapsed.postchange_data.get(field.ct_field)
                            or collapsed.postchange_data.get(field.ct_field + '_id'))
                fk_value = collapsed.postchange_data.get(field.fk_field)
                if ct_value and fk_value == target_obj_id:
                    try:
                        ct = ContentType.objects.get_for_id(ct_value)
                        if ct.model_class() == target_model_class:
                            return True
                    except ContentType.DoesNotExist:
                        pass

        return False

    @staticmethod
    def _split_bidirectional_cycles(collapsed_changes, logger):
        """
        Special case: Preemptively detect and split CREATE operations involved in
        bidirectional FK cycles.

        For each CREATE A with a nullable FK to CREATE B, check if CREATE B has any FK back to A.
        If so, split CREATE A by setting the nullable FK to NULL and creating a separate UPDATE.

        This handles the common case of bidirectional FK relationships (e.g., Circuit ↔ CircuitTermination)
        without needing complex cycle detection.
        """

        creates = {key: c for key, c in collapsed_changes.items() if c.final_action == ActionType.CREATE}
        splits_made = 0

        for key_a, create_a in list(creates.items()):
            if not create_a.postchange_data:
                continue

            # Find nullable FK fields in this CREATE
            for field in create_a.model_class._meta.get_fields():
                if not (isinstance(field, models.ForeignKey) and field.null):
                    continue

                fk_value = create_a.postchange_data.get(field.name)
                if not fk_value:
                    continue

                # Get the target's key
                related_model = field.related_model
                target_ct = ContentType.objects.get_for_model(related_model)
                app_label, model = target_ct.natural_key()
                model_label = f"{app_label}.{model}"
                key_b = (model_label, fk_value)

                # Is target also being created?
                if key_b not in creates:
                    continue

                create_b = creates[key_b]

                # Does target have FK back to us? (bidirectional cycle)
                if SquashMergeStrategy._has_fk_to(create_b, create_a.model_class, key_a[1]):
                    logger.info(
                        f"  Detected bidirectional cycle: {create_a.model_class.__name__}:{key_a[1]} ↔ "
                        f"{create_b.model_class.__name__}:{key_b[1]} (via {field.name})"
                    )

                    # Split create_a: set nullable FK to NULL, create UPDATE to set it later
                    original_postchange = dict(create_a.postchange_data)
                    create_a.postchange_data[field.name] = None

                    # Create UPDATE operation
                    update_key = (key_a[0], key_a[1], f'update_{field.name}')
                    update_collapsed = CollapsedChange(update_key, create_a.model_class)
                    update_collapsed.change_count = 1  # Synthetic update from split
                    update_collapsed.final_action = ActionType.UPDATE
                    update_collapsed.prechange_data = dict(create_a.postchange_data)
                    update_collapsed.postchange_data = original_postchange
                    update_collapsed.last_change = create_a.last_change

                    # Add UPDATE to collapsed_changes
                    collapsed_changes[update_key] = update_collapsed
                    splits_made += 1

                    break

    @staticmethod
    def _log_cycle_details(remaining, collapsed_changes, logger, max_to_show=5):
        """
        Log details about nodes involved in a dependency cycle.
        Used for debugging when cycles are detected.
        """
        for key, deps in list(remaining.items())[:max_to_show]:
            collapsed = collapsed_changes[key]
            model_name = collapsed.model_class.__name__
            obj_id = collapsed.key[1]
            action = collapsed.final_action.upper()

            # Try to get identifying info
            data = collapsed.postchange_data or collapsed.prechange_data or {}
            identifying_info = [f"{field}={data[field]!r}" for field in ['name', 'slug', 'label'] if field in data]
            info_str = f" ({', '.join(identifying_info)})" if identifying_info else ""

            logger.error(f"    {action} {model_name} (ID: {obj_id}){info_str} depends on: {deps}")

        if len(remaining) > max_to_show:
            logger.error(f"    ... and {len(remaining) - max_to_show} more nodes in cycle")

    @staticmethod
    def _dependency_order_by_references(collapsed_changes, logger):
        """
        Orders collapsed changes using topological sort with cycle detection.

        Uses Kahn's algorithm to order nodes respecting their dependency graph.
        Reference: https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm

        The algorithm processes nodes in "layers" - first all nodes with no dependencies,
        then all nodes whose dependencies have been satisfied, and so on.

        When multiple nodes have no dependencies (equal priority in the dependency graph),
        they are ordered by action type priority: DELETE (0) -> UPDATE (1) -> CREATE (2).

        If cycles are detected, raises an exception.

        Returns: ordered list of keys
        """
        logger.info("Adjusting ordering by references...")

        # Define action priority (lower number = higher priority = processed first)
        action_priority = {
            ActionType.DELETE: 0,  # DELETEs should happen first
            ActionType.UPDATE: 1,  # UPDATEs in the middle
            ActionType.CREATE: 2,  # CREATEs should happen last
            ActionType.SKIP: 3,    # SKIPs should never be in the sort, but just in case
        }

        # Create a copy of dependencies to modify
        remaining = {key: set(collapsed.depends_on) for key, collapsed in collapsed_changes.items()}
        ordered = []

        iteration = 0
        max_iterations = len(remaining) * 2  # Safety limit

        while remaining and iteration < max_iterations:
            iteration += 1

            # Find all nodes with no dependencies
            ready = [key for key, deps in remaining.items() if not deps]

            if not ready:
                # No nodes without dependencies - we have a cycle
                logger.error("  Cycle detected in dependency graph.")

                # Log details about the nodes involved in the cycle (for debugging)
                SquashMergeStrategy._log_cycle_details(remaining, collapsed_changes, logger)

                raise Exception(
                    f"Cycle detected in dependency graph. {len(remaining)} changes are involved in "
                    f"circular dependencies and cannot be ordered. This may indicate a complex cycle "
                    f"that could not be automatically resolved. Check the logs above for details."
                )
            # Sort ready nodes by action priority (primary) and time (secondary)
            # This maintains DELETE -> UPDATE -> CREATE ordering, with time ordering within each group
            ready.sort(key=lambda k: (
                action_priority.get(collapsed_changes[k].final_action, 99),
                collapsed_changes[k].last_change.time
            ))

            # Process ready nodes
            for key in ready:
                ordered.append(key)
                del remaining[key]

                # Remove this key from other nodes' dependencies
                for deps in remaining.values():
                    deps.discard(key)

        if iteration >= max_iterations:
            logger.error("  Ordering by references exceeded maximum iterations. Possible complex cycle.")

            # Log details about the remaining unprocessed nodes (for debugging)
            SquashMergeStrategy._log_cycle_details(remaining, collapsed_changes, logger)

            raise Exception(
                f"Ordering by references exceeded maximum iterations ({max_iterations}). "
                f"{len(remaining)} changes could not be ordered, possibly due to a complex cycle. "
                f"Check the logs above for details."
            )

        logger.info(f"  Ordering by references completed: {len(ordered)} changes ordered")
        return ordered

    @staticmethod
    def _order_collapsed_changes(collapsed_changes, logger):
        """
        Order collapsed changes respecting dependencies and time.

        Algorithm:
        1. Initial ordering by time: DELETEs, UPDATEs, CREATEs (each group sorted by time)
        2. Build dependency graph:
           - If UPDATE references deleted object in prechange_data → UPDATE must come before DELETE
             (UPDATE removes the FK reference, allowing the DELETE to proceed)
           - If UPDATE references created object in postchange_data → CREATE must come before UPDATE
             (CREATE must exist before UPDATE can reference it)
           - If CREATE references another created object in postchange_data → referenced CREATE must come first
             (Referenced object must exist before referencing object is created)
        3. Topological sort respecting dependencies

        This ensures:
        - DELETEs generally happen first to free unique constraints
        - UPDATEs that remove FK references happen before their associated DELETEs
        - CREATEs happen before UPDATEs/CREATEs that reference them

        Returns: ordered list of CollapsedChange objects
        """
        logger.info(f"Ordering {len(collapsed_changes)} collapsed changes...")

        # Remove skipped objects
        to_process = {}
        skipped = []
        for k, v in collapsed_changes.items():
            if v.final_action == ActionType.SKIP:
                skipped.append(v)
            else:
                to_process[k] = v

        logger.info(f"  {len(skipped)} changes will be skipped (created and deleted in branch)")
        logger.info(f"  {len(to_process)} changes to process")

        if not to_process:
            return []

        # Preemptively detect and split bidirectional FK cycles
        # This may add new UPDATE operations to to_process
        SquashMergeStrategy._split_bidirectional_cycles(to_process, logger)

        # Group by action and sort each group by time - need this to build the
        # dependency graph correctly
        deletes = sorted(
            [v for v in to_process.values() if v.final_action == ActionType.DELETE],
            key=lambda c: c.last_change.time
        )
        updates = sorted(
            [v for v in to_process.values() if v.final_action == ActionType.UPDATE],
            key=lambda c: c.last_change.time
        )
        creates = sorted(
            [v for v in to_process.values() if v.final_action == ActionType.CREATE],
            key=lambda c: c.last_change.time
        )

        SquashMergeStrategy._build_fk_dependency_graph(deletes, updates, creates, logger)
        ordered_keys = SquashMergeStrategy._dependency_order_by_references(to_process, logger)

        # Convert keys back to collapsed changes
        result = [to_process[key] for key in ordered_keys]
        return result



from netbox.forms import NetBoxModelForm
from utilities.forms.fields import CommentField
from utilities.forms.rendering import FieldSet

from netbox_branching.models import Branch

__all__ = (
    'BranchForm',
)


class BranchForm(NetBoxModelForm):
    fieldsets = (
        FieldSet('name', 'description', 'tags'),
    )
    comments = CommentField()

    class Meta:
        model = Branch
        fields = ('name', 'description', 'comments', 'tags')



"""
Iterative merge strategy implementation.
"""
from django.db import DEFAULT_DB_ALIAS
from netbox.context_managers import event_tracking

from .strategy import MergeStrategy

__all__ = (
    'IterativeMergeStrategy',
)


class IterativeMergeStrategy(MergeStrategy):
    """
    Iterative merge strategy that applies/reverts changes one at a time in chronological order.
    """

    def merge(self, branch, changes, request, logger, user):
        """
        Apply changes iteratively in chronological order.
        """
        models = set()

        for change in changes:
            models.add(change.changed_object_type.model_class())
            with event_tracking(request):
                request.id = change.request_id
                request.user = change.user
                change.apply(branch, using=DEFAULT_DB_ALIAS, logger=logger)

        self._clean(models)

    def revert(self, branch, changes, request, logger, user):
        """
        Undo changes iteratively (one at a time) in reverse chronological order.
        """
        models = set()

        # Undo each change from the Branch
        for change in changes:
            models.add(change.changed_object_type.model_class())
            with event_tracking(request):
                request.id = change.request_id
                request.user = change.user
                change.undo(branch, logger=logger)

        # Perform cleanup tasks
        self._clean(models)



from django.db import connection, migrations
from netbox.plugins import get_plugin_config

from netbox_branching.choices import BranchStatusChoices


def copy_table(apps, schema_editor):
    """
    Create a copy of the extras_tag_object_types table in each active branch.
    """
    Branch = apps.get_model('netbox_branching', 'Branch')

    table = 'extras_tag_object_types'
    schema_prefix = get_plugin_config('netbox_branching', 'schema_prefix')

    with connection.cursor() as cursor:
        main_table = f'public.{table}'

        for branch in Branch.objects.filter(status=BranchStatusChoices.READY):
            print(f'\n    Copying {table} for branch {branch.name} ({branch.schema_id})... ', end='')
            schema_name = f'{schema_prefix}{branch.schema_id}'
            schema_table = f'{schema_name}.{table}'

            # Abort if the table already exists (somehow)
            cursor.execute(
                f"SELECT EXISTS (SELECT 1 FROM pg_tables WHERE schemaname='{schema_name}' AND tablename='{table}')"
            )
            if cursor.fetchone()[0]:
                print('Skipping; table already exists.', end='')
                continue

            # Copy the extras_tag_object_types table to the branch schema
            cursor.execute(f"CREATE TABLE {schema_table} ( LIKE {main_table} INCLUDING INDEXES )")

            # Copy table data
            cursor.execute(f"INSERT INTO {schema_table} SELECT * FROM {main_table}")

            # Set the default value for the ID column to the sequence associated with the source table
            cursor.execute(
                f"ALTER TABLE {schema_table} ALTER COLUMN id SET DEFAULT nextval('extras_tag_object_types_id_seq')"
            )

            # Rename indexes
            cursor.execute(
                f"ALTER INDEX {schema_name}.extras_tag_object_types_objecttype_id_idx "
                f"RENAME TO extras_tag_object_types_contenttype_id_c1b220c3"
            )
            cursor.execute(
                f"ALTER INDEX {schema_name}.extras_tag_object_types_tag_id_idx "
                f"RENAME TO extras_tag_object_types_tag_id_2e1aab29"
            )
            cursor.execute(
                f"ALTER INDEX {schema_name}.extras_tag_object_types_tag_id_objecttype_id_key "
                f"RENAME TO extras_tag_object_types_tag_id_contenttype_id_2ff9910c_uniq"
            )

            print('Success.', end='')

    print('\n ', end='')  # Padding for final "OK"


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_branching', '0005_branch_applied_migrations'),
    ]

    operations = [
        migrations.RunPython(
            code=copy_table,
            reverse_code=migrations.RunPython.noop
        ),
    ]



from extras.webhooks import register_webhook_callback

from netbox_branching.utilities import get_active_branch


@register_webhook_callback
def set_active_branch(object_type, event_type, data, request):
    if request is None:
        return None
    if branch := get_active_branch(request):
        attrs = {
            'id': branch.pk,
            'name': branch.name,
            'schema_id': branch.schema_id,
        }
    else:
        attrs = None
    return {
        'active_branch': attrs,
    }
